# All error types described in DB API 2 are implemented the same way as in
# Django (1.11 to 3.0)., otherwise some exceptions are not correctly reported in it.
from importlib import import_module
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Type, Union
import json
import warnings

import requests  # noqa


# === Forward defs  (they are first due to dependency)

class FakeReq:
    """A Fake Request is used for compatible error reporting in "composite" subrequests."""
    def __init__(self,
                 method: str,
                 url: str,
                 data: str,        # ?? Union[str, List[Any], Dict[str, Any]],
                 headers: Optional[Dict[str, str]] = None,
                 context: Optional[Dict[Any, Any]] = None
                 ) -> None:
        self.method = method
        self.url = url
        self.data = data
        self.headers = headers or {}  # type: Dict[str, str]
        self.context = context or {}  # type: Dict[Any, Any]  # the key is Union[str, int]

    @property
    def body(self) -> str:
        if isinstance(self.data, str):
            return self.data
        return json.dumps(self.data)


class FakeResp:  # pylint:disable=too-many-instance-attributes
    """A Fake Response is used for compatible error reporting in "composite" subrequests."""
    def __init__(self, status_code: int, headers: Mapping[str, str], text: str, request: FakeReq) -> None:
        self.status_code = status_code
        self.text = text
        self.request = request
        self.headers = headers

        self.reason = None


GenResponse = requests.Response  # (requests.Response, 'FakeResp')


# === Exception defs

class SalesforceWarning(Warning):
    def __init__(self,
                 messages: Optional[Union[str, List[str]]] = None,
                 response: Optional[GenResponse] = None,
                 verbs: Optional[Iterable[str]] = None
                 ) -> None:
        self.data = []  # type: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]]
        self.response = None  # type: Optional[GenResponse]
        self.verbs = None  # type: Optional[Set[str]]
        message = prepare_exception(self, messages, response, verbs)
        super().__init__(message)


class Error(Exception):
    """
    Database error that can get detailed error information from a SF REST API response.

    customized for aproriate information, not too much or too little.
    """
    def __init__(self,
                 messages: Optional[Union[str, List[str]]] = None,
                 response: Optional[GenResponse] = None,
                 verbs: Optional[Iterable[str]] = None
                 ) -> None:
        self.data = []  # type: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]]
        self.response = None  # type: Optional[GenResponse]
        self.verbs = None  # type: Optional[Set[str]]
        message = prepare_exception(self, messages, response, verbs)
        super().__init__(message)


class InterfaceError(Error):
    pass  # should be raised directly


class DatabaseError(Error):
    pass


class SalesforceError(DatabaseError):
    """Error reported by SFDC data instance in a REST API request.

    This class is for messages with ExceptionCode that can be searched in
    "SOAP API documentation"
    https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_concepts_core_data_objects.htm#exception_code_topic
    by the capitalized ExceptionCode: e.g. "DUPLICATE_VALUE \n some context about it"

    Subclasses are for messages created by django-salesforce.

    Timeout is also reported as a SalesforceError, because it can be frequently caused
    by SFDC by a slow query. There is no ambiguity.
    """


class DataError(SalesforceError):
    pass


class OperationalError(SalesforceError):
    pass  # e.g. network, auth


class IntegrityError(SalesforceError):
    pass  # e.g. foreign key (probably recently deleted obj)


class InternalError(SalesforceError):
    pass


class ProgrammingError(SalesforceError):
    pass  # e.g sql syntax


class NotSupportedError(SalesforceError):
    pass


class SalesforceAuthError(SalesforceError):
    """Error reported by SFDC in salesforce.auth at login request.

    The messages are typically very cryptic. (probably intentionally,
    to not disclosure more information to unathorized persons)

    Repeated errors of this class can lock the user account temporarily.
    """


class LicenseError(OperationalError):
    pass


def prepare_exception(obj: Union[Error, SalesforceWarning],
                      messages: Optional[Union[str, List[str]]] = None,
                      response: Optional[GenResponse] = None,
                      verbs: Optional[Iterable[str]] = None
                      ) -> str:
    """Prepare excetion params or only an exception message

    parameters:
        messages: list of strings, that will be separated by new line
        response: response from a request to SFDC REST API
        verbs: list of options about verbosity
    """
    # pylint:disable=too-many-branches
    verbs_ = set(verbs or [])
    known_options = ['method+url']
    if messages is None:
        messages = []
    if isinstance(messages, str):
        messages = [messages]
    assert isinstance(messages, list)
    assert not verbs_.difference(known_options)

    data = None
    # a boolean from a failed response is False, though error messages in json should be decoded
    if response is not None and 'json' in response.headers.get('Content-Type', '') and response.text:
        data = json.loads(response.text)
        if data:
            data_0 = data[0]
            if 'errorCode' in data_0:
                subreq = ''
                if 'referenceId' in data_0:
                    subreq = " (in subrequest {!r})".format(data_0['referenceId'])
                messages = [data_0['errorCode'] + subreq] + messages
            if data_0.get('fields'):
                messages.append('FIELDS: {}'.format(data_0['fields']))
            if len(data) > 1:
                messages.append('MORE_ERRORS ({})'.format(len(data)))
    if 'method+url' in verbs_:
        assert response is not None and response.request.url
        method = response.request.method
        url = response.request.url
        if len(url) > 100:
            url = url[:100] + '...'
        data_info = ''
        if (method in ('POST', 'PATCH') and
                (not response.request.body or 'json' not in response.request.headers.get('content-type', ''))):
            data_info = ' (without json request data)'
        messages.append('in {} "{}"{}'.format(method, url, data_info))
    separ = '\n    '
    messages = [x.replace('\n', separ) for x in messages]
    message = separ.join(messages)
    if obj:
        obj.data = data
        obj.response = response
        obj.verbs = verbs_
    return message


def warn_sf(messages: Union[str, List[str]],
            response: Optional[GenResponse],
            verbs: Optional[Iterable[str]] = None,
            klass: Type[SalesforceWarning] = SalesforceWarning
            ) -> None:
    """Issue a warning SalesforceWarning, with message combined from message and data from SFDC response"""
    warnings.warn(klass(messages, response, verbs), stacklevel=2)


def import_string(dotted_path: str) -> Any:
    # copied from django.utils.module_loading
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        ) from err



# Django settings for testrunner project.
from typing import Any, Dict, Optional, Tuple, Union
import os

DEBUG = True

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

SF_EXAMPLE_PERSON_ACCOUNT_ACTIVATED = False

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'salesforce_testrunner_db',
    },
    # The variable DATABASES should be redefined in local_settings with details
    # in order to protect private secret values from unintentional committing.
    'salesforce': {
        'ENGINE': 'salesforce.backend',
        "CONSUMER_KEY": os.environ.get('SF_CONSUMER_KEY', ''),
        "CONSUMER_SECRET": os.environ.get('SF_CONSUMER_SECRET', ''),
        'USER': os.environ.get('SF_USER', ''),
        'PASSWORD': os.environ.get('SF_PASSWORD', ''),
        'HOST': os.environ.get('SF_HOST', 'https://login.salesforce.com'),
        'TEST': {
            'DEPENDENCIES': [],
            'MIGRATE': False,   # to run tests without migrations in Django 3.1+
        },
    }
}  # type: Dict[str, Any]

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/New_York'
USE_TZ = True

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale
# USE_L10N = True  # it is deprecated since Django 4.0 and expected True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    # 'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '6$y&o(28l)#o1_2rafojb_&zxi*jnivkv)ygj#!01kt0ypsxe$'  # used only in tests

SITE_ID = 1

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'salesforce.testrunner.urls'  # type: Optional[str]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
            'debug': DEBUG,
        },
    },
]

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'django.contrib.admindocs',
    'salesforce',
    'salesforce.testrunner.example',
]

SALESFORCE_DB_ALIAS = 'salesforce'
# Timeouts tuple: (waiting for connection, waiting for data) in seconds
SALESFORCE_QUERY_TIMEOUT = (4, 15)  # type: Union[float, Tuple[float, float]]
# Maximal number of retries after timeout.
# REQUESTS_MAX_RETRIES = 1
DATABASE_ROUTERS = [
    "salesforce.router.ModelRouter"
]

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
        },
        'mail_admins': {
            'level': 'ERROR',
            'class': 'django.utils.log.AdminEmailHandler',
            'filters': ['require_debug_false'],
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        'salesforce': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
        'salesforce.testrunner': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
    },
    'filters': {
        'require_debug_false': {
            "()": "django.utils.log.RequireDebugFalse",
        }
    }
}

# Name of primary key - by default 'id'. The value 'Id' was the default for
# version "django-salesforce < 0.5".
# The value SF_PK can not be changed after any migration for Salesforce has been created
# SF_PK = 'Id'

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'  # not important for Salesforce, but for Django warnings

# Django Debug Toolbar >= 4.4.0
# Django changes the DEBUG setting to False when running tests.
# By default the Django Debug Toolbar is installed because DEBUG is set to True.
# For most cases, you need to avoid installing the toolbar when running tests.
# If you feel this check is in error, you can set
# `DEBUG_TOOLBAR_CONFIG['IS_RUNNING_TESTS'] = False` to bypass this check.
DEBUG_TOOLBAR_CONFIG = {'IS_RUNNING_TESTS': False}

DJSF_LICENSE_KEY = os.environ.get('DJSF_LICENSE_KEY', '')  # configure for enterprise features
try:
    from salesforce.testrunner.local_settings import *  # noqa pylint:disable=unused-wildcard-import,wildcard-import
except ImportError:
    pass
